package com.barrelsofdata.examples.beam.streaming.sensor.function;

import com.barrelsofdata.examples.beam.streaming.sensor.model.ComputedEvent;
import com.barrelsofdata.examples.beam.streaming.sensor.model.RawEvent;
import org.apache.beam.sdk.transforms.Combine;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.StreamSupport;

/**
 * Combiner function that emits a {@link ComputedEvent} by computing averages of values
 * from a group of {@link RawEvent}
 */
public class Average extends Combine.CombineFn<RawEvent, ArrayList<RawEvent>, ComputedEvent> {
    @Override
    public ArrayList<RawEvent> createAccumulator() {
        // Initializes an accumulator at worker level
        return new ArrayList<>();
    }

    @Override
    public ArrayList<RawEvent> addInput(ArrayList<RawEvent> accumulator, RawEvent input) {
        // Adds data into accumulator, at worker level
        accumulator.add(input);
        return accumulator;
    }

    @Override
    public ArrayList<RawEvent> mergeAccumulators(Iterable<ArrayList<RawEvent>> accumulators) {
        // Merging accumulators from all workers into one
        List<RawEvent> flattened = StreamSupport.stream(accumulators.spliterator(), true).flatMap(ArrayList::stream).toList();
        return new ArrayList<>(flattened);
    }

    @Override
    public ComputedEvent extractOutput(ArrayList<RawEvent> accumulator) {
        // Aggregating and emitting one result

        // Get id and timestamp of last event to use as timestamp of the new event
        // Getting id here make it easy for us to throwaway the grouping keys later
        Collections.sort(accumulator);
        RawEvent lastRawEventInWindow = accumulator.get(accumulator.size() - 1);
        String id = lastRawEventInWindow.id();
        Long timestamp = lastRawEventInWindow.data().ts();

        Double average = accumulator.stream()
                .mapToDouble(ev -> ev.data().value())
                .average()
                .orElse(-1);

        return new ComputedEvent(id, timestamp, average);
    }
}
