package com.barrelsofdata.examples.beam.streaming.sensor.pipeline;

import com.barrelsofdata.examples.beam.streaming.sensor.config.SensorPipelineOptions;
import com.barrelsofdata.examples.beam.streaming.sensor.config.Steps;
import com.barrelsofdata.examples.beam.streaming.sensor.model.ComputedEvent;
import com.barrelsofdata.examples.beam.streaming.sensor.model.RawEvent;
import com.barrelsofdata.examples.beam.streaming.sensor.target.JDBCTarget;
import com.barrelsofdata.examples.beam.streaming.sensor.transform.ComputeAverage;
import com.barrelsofdata.examples.beam.streaming.sensor.source.KafkaSource;
import com.barrelsofdata.examples.beam.streaming.sensor.transform.ParseWithTimestamp;
import org.apache.beam.sdk.Pipeline;
import org.apache.beam.sdk.values.KV;
import org.apache.beam.sdk.values.PCollection;
import org.apache.beam.sdk.values.PDone;
import org.joda.time.Duration;

/**
 * This class builds the sensor pipeline by applying various transforms
 */
public class SensorAnalytics {

    private static final Long WINDOW_DURATION_SEC = 180L; // 3 minutes
    private static final Long WINDOW_FREQUENCY_SEC = 60L; // Every 1 minute
    private static final Long ALLOWED_LATENESS_SEC = 120L; // 2 minutes

    private SensorAnalytics() {};

    /**
     * Builds the pipeline by consuming the passed pipeline options
     * @param pipeline Pipeline object to apply the transforms to
     * @param options Configuration to be used for the pipeline
     */
    public static Pipeline from(Pipeline pipeline, SensorPipelineOptions options) {

        KafkaSource kafkaSource = new KafkaSource(
                options.getKafkaBrokers(),
                options.getKafkaTopic(),
                options.getKafkaConsumerGroupId(),
                options.isResetToEarliest());

        // Convert o bounded source - Only for testing
        Integer kafkaMaxReadDuration = options.getKafkaMaxReadDurationMinutes();
        if(kafkaMaxReadDuration != null)
            kafkaSource.withMaxReadTime(Duration.standardMinutes(kafkaMaxReadDuration));

        JDBCTarget jdbcTarget = new JDBCTarget(
                options.getSqlDriver(),
                options.getJdbcUrl(),
                options.getTable(),
                options.getUsername(),
                options.getPassword());

        PCollection<KV<String, String>> readFromSource = pipeline.apply(Steps.SOURCE.getStep(),
                kafkaSource);
        PCollection<RawEvent> parsed = readFromSource.apply(Steps.PARSE.getStep(),
                new ParseWithTimestamp());
        PCollection<ComputedEvent> averaged = parsed.apply(Steps.PROCESS_DATA.getStep(),
                new ComputeAverage(WINDOW_DURATION_SEC, WINDOW_FREQUENCY_SEC, ALLOWED_LATENESS_SEC));
        PDone writeToTarget = averaged.apply(Steps.TARGET.getStep(),
                jdbcTarget);

        return writeToTarget.getPipeline();
    }
}
