package com.barrelsofdata.examples.beam.streaming.sensor.source;

import com.barrelsofdata.examples.beam.streaming.sensor.config.Steps;
import org.apache.beam.sdk.Pipeline;
import org.apache.beam.sdk.io.kafka.KafkaIO;
import org.apache.beam.sdk.transforms.PTransform;
import org.apache.beam.sdk.values.KV;
import org.apache.beam.sdk.values.PCollection;
import org.apache.beam.sdk.values.PInput;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.joda.time.Duration;

import java.util.Map;

/**
 * PTransform that connects to kafka and reads from the given topic
 */
public class KafkaSource extends PTransform<PInput, PCollection<KV<String, String>>> {

    private final String brokers;
    private final String topic;
    private final String consumerGroup;
    private final Boolean resetToEarliest;

    private Duration maxReadTime;

    public KafkaSource(String brokers, String topic, String consumerGroup, Boolean resetToEarliest) {
        this.brokers = brokers;
        this.topic = topic;
        this.consumerGroup = consumerGroup;
        this.resetToEarliest = resetToEarliest;
    }

    /**
     * Sets the amount of time to keep reading from kafka
     * Only used in testing to convert unbounded source to bounded source
     * @param maxReadTime Duration to read data
     */
    public void withMaxReadTime(Duration maxReadTime) {
        this.maxReadTime = maxReadTime;
    }

    @Override
    public PCollection<KV<String, String>> expand(PInput input) {
        Pipeline pipeline = input.getPipeline();
        KafkaIO.Read<String, String> kafkaIo = KafkaIO.<String, String>read()
                .withBootstrapServers(brokers)
                .withTopic(topic)
                .withKeyDeserializer(StringDeserializer.class)
                .withValueDeserializer(StringDeserializer.class)
                .withConsumerConfigUpdates(Map.of(
                        "group.id", consumerGroup,
                        "auto.offset.reset", resetToEarliest ? "earliest" : "latest"))
                .commitOffsetsInFinalize()
                .withLogAppendTime(); // Do not use timestamp from the event yet

        if(maxReadTime != null)
            kafkaIo = kafkaIo.withMaxReadTime(maxReadTime);

        return pipeline.apply(Steps.READ_KAFKA.getStep(), kafkaIo.withoutMetadata());
    }
}
